"""
gnusocial.utils
~~~~~~~~~~~~~~~

Module with various utility functions and exception classes.
"""
import re
from typing import Callable
from functools import partial
import requests
from requests.auth import HTTPBasicAuth
from dtd import docstring
from .docs import _SERVER_URL_DOC, _CONFIG_DICT, _USERNAME_DOC, _PASSWORD_DOC

DOMAIN_REGEX = re.compile(r"http(s|)://(www\.|)(.+?)(/.*|)$")


@docstring(server_url=_SERVER_URL_DOC)
class ServerURLError(Exception):
    """Exception class for errors in server URL.

:param str server_url: {server_url}
    """
    def __init__(self, server_url: str) -> None:
        self.server_url = server_url
        super().__init__(Exception)

    def __repr__(self) -> str:
        return 'ServerURLError(%r)' % self.server_url

    def __str__(self) -> str:
        return 'Invalid server URL %s' % self.server_url


@docstring(server_url=_SERVER_URL_DOC,
           username=_USERNAME_DOC,
           password=_PASSWORD_DOC)
class AuthenticationError(Exception):
    """Exception class for authentication errors.

:param str server_url: {server_url}
:param str username: {username}
:param str password: {password}
    """
    def __init__(self, server_url: str, username: str, password: str) -> None:
        self.server_url = server_url
        self.username = username
        self.password = password
        super().__init__(self)

    def __repr__(self) -> str:
        return 'AuthenticationError(%r, %r, %r)' % (self.server_url,
                                                    self.username,
                                                    self.password)

    def __str__(self) -> str:
        return 'Invalid credentials for (%s, %s)' % (self.username,
                                                     self.server_url)


def _check_auth_error(response: requests.models.Response,
                      server_url: str,
                      username: str,
                      password: str) -> None:
    if response.status_code == 401:
        raise AuthenticationError(server_url, username, password)


def _api_path(server_url: str) -> str:
    _validate_server_url(server_url)
    if server_url[-1] != '/':
        server_url += '/'
    return server_url + 'api/'


def _validate_server_url(server_url: str) -> None:
    if not DOMAIN_REGEX.match(server_url):
        raise ServerURLError(server_url)


def _check_connection(server_url: str) -> None:
    response = _get_request(server_url, 'help/test').json()
    if response != 'ok':
        raise requests.ConnectionError(server_url)


def _resource_url(server_url: str,
                  resource_path: str,
                  extension: str='.json') -> str:
    return _api_path(server_url) + resource_path + extension


def _request(request_func: Callable,
             server_url: str,
             resource_path: str,
             **kwargs) -> requests.models.Response:
    extension = kwargs.get('extension', '.json')
    req = partial(request_func,
                  url=_resource_url(server_url, resource_path, extension),
                  data=kwargs.get('data'),
                  files=kwargs.get('media'),
                  params=kwargs.get('params'))
    response = None
    username = kwargs.get('username')
    oauth = kwargs.get('oauth')
    if username and oauth:
        raise Exception("You can't use HTTP Basic Auth and OAuth at the same" +
                        "time.")
    if username:
        password = kwargs.get('password')
        response = req(auth=HTTPBasicAuth(username, password))
        _check_auth_error(response, server_url, username, password)
    elif oauth:
        response = req(auth=oauth)
    else:
        response = req()
    return response


def _get_request(server_url: str,
                 resource_path: str,
                 **kwargs) -> requests.models.Response:
    return _request(request_func=requests.get,
                    server_url=server_url,
                    resource_path=resource_path,
                    **kwargs)


def _post_request(server_url: str,
                  resource_path: str,
                  **kwargs) -> requests.models.Response:
    return _request(request_func=requests.post,
                    server_url=server_url,
                    resource_path=resource_path,
                    **kwargs)


@docstring(server_url=_SERVER_URL_DOC, config_dict=_CONFIG_DICT)
def config(server_url: str) -> dict:
    """Returns server configuration.

:param str server_url: {server_url}
:rtype: dict
:return: dict with following structure:

::

    {config_dict}

    """
    return _get_request(server_url, 'statusnet/config').json()


def _check_user_target(username: str='', **kwargs) -> None:
    both_targets = 'user_id' in kwargs and 'screen_name' in kwargs
    no_targets = 'user_id' not in kwargs and 'screen_name' not in kwargs
    if both_targets:
        raise Exception(
            "You must either specify the user_id or screen_name."
        )
    if no_targets and not username:
        raise Exception(
            "You must either specify the user_id or screen_name or " +
            "username."
        )


def _check_id_and_nickname(**kwargs) -> None:
    has_group_id = 'id' in kwargs
    has_group_name = 'nickname' in kwargs
    if has_group_id == has_group_name:
        raise Exception(
            "You must either specify the id or nickname."
        )


def _check_group_id_and_name(**kwargs) -> None:
    has_group_id = 'group_id' in kwargs
    has_group_name = 'group_name' in kwargs
    if has_group_id == has_group_name:
        raise Exception(
            "You must either specify the group_id or group_name."
        )



"""
gnusocial.oauth
~~~~~~~~~~~~~~~

Module with OAuth resources.
"""
from urllib.parse import parse_qs
from dtd import docstring
from requests_oauthlib import OAuth1
from .utils import _post_request, _resource_url
from .docs import _SERVER_URL_DOC

_CONSUMER_KEY = """A value used by the application to identify itself to the GNU
    Social server"""
_CONSUMER_SECRET = """A secret used by the application to establish ownership of
    the consumer key"""
_RESOURCE_OWNER_KEY = """A key used by user to authorize"""
_RESOURCE_OWNER_SECRET = """A secret used by the application to establish
    ownership of a given token"""
_RESOURCE_OWNER_DICT = """dict with 'resource_owner_key' and
    'resource_owner_secret' keys"""


def _parse_response(response) -> dict:
    creds = parse_qs(response.text)
    res = {}
    res['resource_owner_key'] = creds['oauth_token'][0]
    res['resource_owner_secret'] = creds['oauth_token_secret'][0]
    return res


@docstring(server_url=_SERVER_URL_DOC,
           consumer_key=_CONSUMER_KEY,
           consumer_secret=_CONSUMER_SECRET,
           resource_owner_dict=_RESOURCE_OWNER_DICT)
def request_token(server_url: str,
                  consumer_key: str,
                  consumer_secret: str,
                  oauth_callback: str='oob') -> dict:
    """Returns resource owner key and secret for given consumer key and secret.

:param str server_url: {server_url}
:param str consumer_key: {consumer_key}
:param str consumer_secret: {consumer_secret}
:param str oauth_callback: (optional) A callback URL
:rtype: dict
:return: {resource_owner_dict}
    """
    oauth = OAuth1(client_key=consumer_key, client_secret=consumer_secret)
    return _parse_response(_post_request(
        server_url=server_url,
        resource_path='oauth/request_token',
        oauth=oauth,
        data={'oauth_callback': oauth_callback},
        extension=''
    ))


@docstring(server_url=_SERVER_URL_DOC,
           resource_owner_key=_RESOURCE_OWNER_KEY)
def authorize_url(server_url: str,
                  resource_owner_key: str) -> str:
    """Returns a URL used to obtain user authorization for application access.

:param str server_url: {server_url}
:param str resource_owner_key: {resource_owner_key}
:rtype: str
:return: a URL used to obtain user authorization for application access
    """
    return _resource_url(server_url, 'oauth/authorize', extension='') +\
        '?oauth_token=' + resource_owner_key


@docstring(server_url=_SERVER_URL_DOC,
           consumer_key=_CONSUMER_KEY,
           consumer_secret=_CONSUMER_SECRET,
           resource_owner_key=_RESOURCE_OWNER_KEY,
           resource_owner_secret=_RESOURCE_OWNER_SECRET,
           resource_owner_dict=_RESOURCE_OWNER_DICT)
def access_token(server_url: str,
                 consumer_key: str,
                 consumer_secret: str,
                 resource_owner_key: str,
                 resource_owner_secret: str,
                 secret_key: str) -> dict:
    """Returns access key and secret.

:param str server_url: {server_url}
:param str consumer_key: {consumer_key}
:param str consumer_secret: {consumer_secret}
:param str oauth_callback: (optional) A callback URL
:rtype: dict
:return: {resource_owner_dict}
    """
    oauth = OAuth1(client_key=consumer_key,
                   client_secret=consumer_secret,
                   resource_owner_key=resource_owner_key,
                   resource_owner_secret=resource_owner_secret,
                   verifier=secret_key)
    return _parse_response(_post_request(
        server_url=server_url,
        resource_path='oauth/access_token',
        oauth=oauth
    ))



"""
gnusocial.activity_streams
~~~~~~~~~~~~~~~~~~~~~~~~~~

Module with ActivityStream timeline resources
"""
from string import Template
from dtd import docstring
from .utils import _get_request, _check_user_target
from .docs import (_SERVER_URL_DOC, _USERNAME_DOC, _PASSWORD_DOC, _MAX_ID_DOC,
                    _SINCE_ID_DOC, _STATUSES_COUNT, _USER_ID_DOC,
                    _SCREEN_NAME_DOC)


_TAG = '\'tag:gs.smuglo.li,2016-11-16:noticeId=1028155:objectType=comment\''

_TAG_EXAMPLE = 'Example: {tag}'.format(tag=_TAG)

_CONVERSATION_TAG = 'tag:social.heldscal.la,2016-11-15:objectType=thread:' +\
                    'nonce=c53d340f4850ca73'
# TODO: add more verbs
_AS_DICT = Template("""{
        'generator': str, # string describing the server software
        'totalItems': int,
        'title': str, # string describing the result dict
        'links':
        [
            {
                'rel': 'alternate',
                'type': 'text/html',
                'url': str # URL with corresponding content
            }
        ],
        'items':
        [
            {
                'actor':
                {
                    'displayName': str,
                    'id': str, # URL to profile
                    'image':
                    {
                        'height': int,
                        'rel': 'avatar',
                        'type': str, # MIME type
                        'url': str,
                        'width': int
                    },
                    'objectType': 'person',
                    'portablecontacts_net':
                    {
                        'displayName': str,
                        'note': str, # user profile description
                        'preferredUsername': str, # screen name
                        'urls':
                        [
                            {
                                'primary': 'true',
                                'type': 'homepage',
                                'value': str, # URL to the homepage
                            }
                        ]
                    },
                    'status_net':
                    {
                        'avatarLinks':
                        [
                            {
                                'height': int,
                                'rel': 'avatar',
                                'type': str, # MIME type
                                'url': str,
                                'width': int
                            }
                        ],
                        'profile_info':
                        {
                            'local_id': str, # ID of the user on their home server
                        }
                    },
                    'summary': str, # user profile description
                    'url': str,
                },
                'content': str, # status in HTML
                'generator':
                {
                    'id': str, # $tag_example
                    'objectType': 'application',
                    'status_net':
                    {
                        'source_code': 'ostatus'
                    }
                },
                'id': str, #
                'object':
                {
                    'content': str, # status in HTML
                    'id': str, # $tag_example
                    'inReplyTo':
                    {
                        'id': str, # $tag_example
                        'objectType': 'note',
                        'url': str,
                    },
                    'objectType': 'comment',
                    'status_net':
                    {
                        'notice_id': int,
                    },
                    'url': str,
                },
                'provider':
                {
                    'displayName': str,
                    'objectType': 'service',
                    'url': str,
                },
                'published': str, # Example: '2016-11-16T20:42:18+00:00',
                'status_net':
                {
                    'conversation': str, # Example: $conversation_tag,
                    'notice_info':
                    {
                        'local_id': str,
                        'source': 'ostatus'
                    }
                },
                'to': # users, to whom the reply is addressed
                [
                    {
                        'id': str, # URL to profile
                        'objectType': 'http://activitystrea.ms/schema/1.0/person'
                    },
                    {
                        'id': 'http://activityschema.org/collection/public',
                        'objectType': 'http://activitystrea.ms/schema/1.0/collection'
                    }
                ],
                'url': str,
                'verb': str, # Example: 'post'
            }
        ]
    }
""")
_AS_DICT = _AS_DICT.substitute(tag_example=_TAG_EXAMPLE,
                               conversation_tag=_CONVERSATION_TAG)


@docstring(server_url=_SERVER_URL_DOC,
           username=_USERNAME_DOC,
           password=_PASSWORD_DOC,
           since_id=_SINCE_ID_DOC,
           max_id=_MAX_ID_DOC,
           count=_STATUSES_COUNT,
           as_dict=_AS_DICT)
def public(server_url: str,
           username: str='',
           password: str='',
           **kwargs) -> dict:
    """Returns the most recent notices, including repeats if they exist, from
non-protected users.

:param str server_url: {server_url}
:param str username: (optional) {username}
:param str password: (optional) {password}
:param int since_id: (optional) {since_id}
:param int max_id: (optional) {max_id}
:param int count: (optional) {count}
:rtype: dict
:return: dict with following structure:

::

    {as_dict}
    """
    return _get_request(server_url=server_url,
                        resource_path='statuses/public_timeline',
                        extension='.as',
                        username=username,
                        password=password,
                        params=kwargs).json()


@docstring(server_url=_SERVER_URL_DOC,
           username=_USERNAME_DOC,
           password=_PASSWORD_DOC,
           since_id=_SINCE_ID_DOC,
           max_id=_MAX_ID_DOC,
           count=_STATUSES_COUNT,
           as_dict=_AS_DICT)
def home(server_url: str, username: str, password: str, **kwargs) -> list:
    """Returns the most recent notices, including repeats if they exist,
posted by the authenticating user and the users they follow. This is the
same timeline seen by a user when they login to their instance.

:param str server_url: {server_url}
:param str username: (optional) {username}
:param str password: (optional) {password}
:param int since_id: (optional) {since_id}
:param int max_id: (optional) {max_id}
:param int count: (optional) {count}
:rtype: dict
:return: dict with following structure:

::

    {as_dict}
    """
    return _get_request(server_url=server_url,
                        resource_path='statuses/home_timeline',
                        extension='.as',
                        username=username,
                        password=password,
                        params=kwargs).json()


@docstring(server_url=_SERVER_URL_DOC,
           username=_USERNAME_DOC,
           password=_PASSWORD_DOC,
           since_id=_SINCE_ID_DOC,
           max_id=_MAX_ID_DOC,
           count=_STATUSES_COUNT,
           user_id=_USER_ID_DOC,
           screen_name=_SCREEN_NAME_DOC,
           as_dict=_AS_DICT)
def friends(server_url: str,
            username: str='',
            password: str='',
            **kwargs) -> list:
    """Alias of statuses/home_timeline for the specified user.

:param str server_url: {server_url}
:param str username: (optional) {username}
:param str password: (optional) {password}
:param int user_id: (optional) {user_id}
:param str screen_name: (optional) {screen_name}
:param int since_id: (optional) {since_id}
:param int max_id: (optional) {max_id}
:param int count: (optional) {count}
:rtype: dict
:return: dict with following structure:

::

    {as_dict}
    """
    _check_user_target(username, **kwargs)
    return _get_request(server_url=server_url,
                        resource_path='statuses/friends_timeline',
                        extension='.as',
                        username=username,
                        password=password,
                        params=kwargs).json()


@docstring(server_url=_SERVER_URL_DOC,
           username=_USERNAME_DOC,
           password=_PASSWORD_DOC,
           since_id=_SINCE_ID_DOC,
           max_id=_MAX_ID_DOC,
           count=_STATUSES_COUNT,
           user_id=_USER_ID_DOC,
           screen_name=_SCREEN_NAME_DOC,
           as_dict=_AS_DICT)
def user(server_url: str,
         username: str='',
         password: str='',
         **kwargs) -> list:
    """Returns the most recent notices posted by the authenticating user. It
is also possible to request another user's timeline by using the
screen_name or user_id parameter. The other users timeline will only be
visible if they are not protected, or if the authenticating user's follow
request was accepted by the protected user.

:param str server_url: {server_url}
:param str username: (optional) {username}
:param str password: (optional) {password}
:param int user_id: (optional) {user_id}
:param str screen_name: (optional) {screen_name}
:param int since_id: (optional) {since_id}
:param int max_id: (optional) {max_id}
:param int count: (optional) {count}
:rtype: dict
:return: dict with following structure:

::

    {as_dict}
    """
    _check_user_target(username, **kwargs)
    return _get_request(server_url=server_url,
                        resource_path='statuses/user_timeline',
                        extension='.as',
                        username=username,
                        password=password,
                        params=kwargs).json()


@docstring(server_url=_SERVER_URL_DOC,
           username=_USERNAME_DOC,
           password=_PASSWORD_DOC,
           since_id=_SINCE_ID_DOC,
           max_id=_MAX_ID_DOC,
           count=_STATUSES_COUNT,
           user_id=_USER_ID_DOC,
           screen_name=_SCREEN_NAME_DOC,
           as_dict=_AS_DICT)
def mentions(server_url: str,
             username: str='',
             password: str='',
             **kwargs) -> list:
    """Returns the most recent mentions (notices containing @username) for
the authenticating user or specified user.

:param str server_url: {server_url}
:param str username: (optional) {username}
:param str password: (optional) {password}
:param int user_id: (optional) {user_id}
:param str screen_name: (optional) {screen_name}
:param int since_id: (optional) {since_id}
:param int max_id: (optional) {max_id}
:param int count: (optional) {count}
:rtype: dict
:return: dict with following structure:

::

    {as_dict}
    """
    _check_user_target(username, **kwargs)
    return _get_request(server_url=server_url,
                        resource_path='statuses/mentions',
                        extension='.as',
                        username=username,
                        password=password,
                        params=kwargs).json()


@docstring(server_url=_SERVER_URL_DOC,
           username=_USERNAME_DOC,
           password=_PASSWORD_DOC,
           since_id=_SINCE_ID_DOC,
           max_id=_MAX_ID_DOC,
           count=_STATUSES_COUNT,
           user_id=_USER_ID_DOC,
           screen_name=_SCREEN_NAME_DOC,
           as_dict=_AS_DICT)
def replies(server_url: str,
            username: str='',
            password: str='',
            **kwargs) -> list:
    """Alias of statuses/mentions

:param str server_url: {server_url}
:param str username: (optional) {username}
:param str password: (optional) {password}
:param int user_id: (optional) {user_id}
:param str screen_name: (optional) {screen_name}
:param int since_id: (optional) {since_id}
:param int max_id: (optional) {max_id}
:param int count: (optional) {count}
:rtype: dict
:return: dict with following structure:

::

    {as_dict}
    """
    _check_user_target(username, **kwargs)
    return _get_request(server_url=server_url,
                        resource_path='statuses/replies',
                        extension='.as',
                        username=username,
                        password=password,
                        params=kwargs).json()
